import json
from ... import cnscom

def complete_ident(cnsapp, textst):
	ret = []
	statuses = cnsapp.cnsapp.svrcall(cnscom.callid_status, json.dumps({}), auto_connect=True)
	for st in json.loads(statuses):
		if st['ident'].startswith(textst) or textst == "":
			ret.append(st['ident'])
	return ret



import sys, os, socket, errno, logging, time
from ..config import config, read_config, config_files
from ..utils import launch_server_daemonized
from .. import cnscom
from .parser import argparser
from . import exception

###

L = logging.getLogger("cnsapp")

###

class console_app(object):
	'''
Console application (base for custom implementations)
	'''

	def __init__(self, configuration):
		'''
		@param configuration: string or list of configuration files that will be used by Ramona. This is application level configuration.
		'''
		# Change directory to location of user console script
		os.chdir(os.path.dirname(sys.argv[0]))

		# Check if this is request for proxy tool - and avoid parsing
		if len(sys.argv) > 1:
			for mn in dir(self):
				fn = getattr(self, mn)
				if not hasattr(fn, 'proxy_tool'): continue
				if mn == sys.argv[1]:
					ret = fn(sys.argv[1:])
					sys.exit(ret)

		# Parse command line arguments
		self.argparser = argparser(self)

		if (len(sys.argv) < 2):
			# Default command
			argv = ['console']
		else:
			argv = None

		self.argparser.parse(argv)

		# Read config
		if self.argparser.args.config is None:
			if isinstance(configuration, basestring):
				configuration = [configuration]
			else:
				pass
		else:
			configuration = self.argparser.args.config
		for config_file in configuration:
			config_file = config_file.strip()
			if not os.path.isfile(config_file):
				print("Cannot find configuration file {0}".format(config_file))
				sys.exit(exception.configuration_error.exitcode)
		
		try:
			read_config(configuration, use_env=False)
		except Exception, e:
			print("{0}".format(e))
			sys.exit(exception.configuration_error.exitcode)

		# Configure logging
		llvl = logging.INFO
		if self.argparser.args.silent: llvl = logging.ERROR
		if self.argparser.args.debug: llvl = logging.DEBUG
		logging.basicConfig(
			level=llvl,
			stream=sys.stderr,
			format="%(asctime)s %(levelname)s: %(message)s",
		)
		if self.argparser.args.debug:
			L.debug("Debug output is enabled.")

		L.debug("Configuration read from: {0}".format(', '.join(config_files)))

		# Prepare server connection factory
		self.cnsconuri = cnscom.socket_uri(config.get('ramona:console','serveruri'))
		self.ctlconsock = None


	def run(self):
		try:
			ec = self.argparser.execute(self)
		except exception.ramona_runtime_errorbase, e:
			L.error("{0}".format(e))
			ec = e.exitcode
		except KeyboardInterrupt, e:
			ec = 0
		except Exception, e:
			L.error("{0}".format(e))
			ec = 100 # Generic error exit code
		sys.exit(ec if ec is not None else 0)


	def connect(self):
		if self.ctlconsock is None: 
			try:
				self.ctlconsock = self.cnsconuri.create_socket_connect()
			except socket.error, e:
				if e.errno == errno.ECONNREFUSED: return None
				if e.errno == errno.ENOENT and self.cnsconuri.protocol == 'unix': return None
				raise

		return self.ctlconsock


	def svrcall(self, callid, params="", auto_connect=False, auto_server_start=False):
		'''
		@param auto_connect: Automatically establish server connection if not present
		@param auto_server_start: Automatically start server if not running and establish connection
		'''
		assert not (auto_connect & auto_server_start), "Only one of auto_connect and auto_server_start can be true"
		if auto_connect:
			if self.ctlconsock is None:
				s = self.connect()
				if s is None:
					raise exception.server_not_responding_error("Server is not responding - maybe it isn't running.")

		elif auto_server_start:
			# Fist check if ramona server is running and if not, launch that
			s = self.auto_server_start()

		else:
			assert self.ctlconsock is not None

		try:
			return cnscom.svrcall(self.ctlconsock, callid, params)
		except socket.error:
			pass

		if auto_connect or auto_server_start:
			L.debug("Reconnecting to server ...")

			self.ctlconsock = None
			s = self.connect()
			if s is None:
				raise exception.server_not_responding_error("Server is not responding - maybe it isn't running.")

			return cnscom.svrcall(self.ctlconsock, callid, params)


	def wait_for_svrexit(self):
		if self.ctlconsock is None: return
		while True:
			x = self.ctlconsock.recv(4096)
			if len(x) == 0: break
		self.ctlconsock
		self.ctlconsock = None


	def auto_server_start(self):
		s = self.connect()
		if s is None:
			L.debug("It looks like Ramona server is not running - launching server")
			launch_server_daemonized()

			for _ in range(100): # Check server availability for next 10 seconds 
				# TODO: Also improve 'crash-start' detection (to reduce lag when server fails to start)
				time.sleep(0.1)
				s = self.connect()
				if s is not None: break

		if s is None:
			raise exception.server_start_error("Ramona server process start failed")

		return s

###

def tool(fn):
	'''
	Tool decorator foc console_app
	'''
	fn.tool = fn.func_name
	return fn

#

def proxy_tool(fn):
	'''
	Proxy tool (with straight argument passing) decorator foc console_app
	'''
	fn.proxy_tool = fn.func_name
	return fn



import json
from ... import cnscom
from .. import exception
from ._completes import complete_ident

###

name = 'stop'
cmdhelp = 'Terminate subprocess(es)'

###

def init_parser(parser):
	parser.add_argument('program', nargs='*', help='Optionally specify program(s) in scope of the command')
	parser.add_argument('-i','--immediate-return', action='store_true', help='Dont wait for start of subprocesses and return ASAP')
	parser.add_argument('-c','--core-dump', action='store_true', help='Stop and dump core (core dump must be enabled in program configuration)')
	parser.add_argument('-E','--stop-and-exit', action='store_true', help='Stop all programs and exit Ramona server. Command-line default for:\n%(prog)s')
	parser.add_argument('-T','--stop-and-stay', action='store_true', help='Stop all programs and keep server running.')

###

def complete(cnsapp, text, line, begidx, endidx):
        textst = text.strip()
        ret = []
        ret.extend(complete_ident(cnsapp, textst))
        return ret

###

def main(cnsapp, args):
	if args.stop_and_exit and len(args.program) > 0:
		raise exception.parameters_error('Cannot specify programs and -E option at once.')

	if args.stop_and_exit and args.stop_and_stay:
		raise exception.parameters_error('Cannot specify -T and -E option at once.')

	if len(args.program) == 0 and not args.stop_and_stay:
		args.stop_and_exit = True

	params={
		'immediate': args.immediate_return,
		'coredump': args.core_dump,
	}
	if args.stop_and_exit: params['mode'] = 'exit'
	elif args.stop_and_stay: params['mode'] = 'stay'
	if len(args.program) > 0: params['pfilter'] = args.program

	cnsapp.svrcall(
		cnscom.callid_stop,
		json.dumps(params),
		auto_connect=True
	)

	if args.stop_and_exit:
		cnsapp.wait_for_svrexit()



import smtplib, urlparse, logging, socket, getpass
from email.mime.text import MIMEText
from ..config import config

#

L = logging.getLogger('notify')

#

class notificator(object):

	# Configure urlparce
	if 'smtp' not in urlparse.uses_params: urlparse.uses_params.append('smtp')

	def __init__(self, svrapp):
		self.appname = config.get('general','appname')

		delivery = config.get('ramona:notify','delivery').strip()
		if delivery == '':
			self.delivery = None
		else:
			delurl = urlparse.urlparse(delivery)
			if delurl.scheme == 'smtp' :
				if delurl.hostname is None:
					L.error("Delivery URL has no hostname: {0}".format(delivery))
					self.delivery = None
				else:
					self.delivery = delurl
					try: # Connection test
						smtpcon = smtplib.SMTP(self.delivery.hostname, self.delivery.port)
						smtpcon.quit()
					except Exception, e:
						L.warning("Given SMTP server ({1}) is not responding: {0}".format(e, delivery))

					self.sender = config.get('ramona:notify','sender')
					if self.sender == '<user>':
						self.sender = '{0}@{1}'.format(getpass.getuser(), socket.gethostname())
					elif self.sender[:1] == '<':
						L.error('Invalid sender option: {0}'.format(self.sender))
						self.delivery = None

					self.receiver = config.get('ramona:notify', 'receiver').split(',')

		#TODO: <sendmail> - see http://stackoverflow.com/questions/73781/sending-mail-via-sendmail-from-python
		#TODO: cmd:custom.sh


	def publish(self, target, prog_ident, stream_name, pattern, tail):

		hostname = socket.gethostname()

		nfttext = 'Application: {0}\n'.format(self.appname)
		nfttext += 'Program: {0}\n'.format(prog_ident)
		nfttext += 'Pattern: {0}\n'.format(pattern)
		nfttext += 'Hostname: {0}\n'.format(hostname)
		nfttext += '\n'+'-'*80+'\n'
		nfttext += tail	
		nfttext += '\n'+'-'*80+'\n'

		if target.startswith('mailto'):
			receiver = urlparse.urlparse(target).path
		else: 
			receiver = None

		#TODO: Decide what to do based on 'target' value

		self._send_mail('{0} / {1} / {2} / {3}'.format(
				self.appname,
				prog_ident,
				pattern,
				hostname,
				), 
			nfttext,
			receiver
		)


	def _send_mail(self, subject, text, receiver=None):
		try:
			text = ''.join([
				'Hello,\n\nRamona detected following condition:\n',
				text,
				'\nBest regards,\nYour Ramona\n\nhttp://ateska.github.com/ramona\n'
			])

			sender = self.sender
			if receiver is None:
				receiver = self.receiver
			elif isinstance(receiver, basestring):
				receiver = [receiver]


			msg = MIMEText(text, 'plain', 'utf-8')
			msg['Subject'] = subject + ' by Ramona'
			msg['From'] = sender
			msg['To'] = ', '.join(receiver)

			s = smtplib.SMTP(self.delivery.hostname, self.delivery.port)
			p = dict(urlparse.parse_qsl(self.delivery.params))
			if p.get('tls', 1): s.starttls()
			if self.delivery.username is not None and self.delivery.password is not None:
				s.login(self.delivery.username, self.delivery.password)
			s.sendmail(sender, receiver, msg.as_string())
			s.quit()

		except:
			L.exception('Exception during sending mail - ignoring')



class ramona_runtime_errorbase(RuntimeError):
	exitcode = 100

class server_not_responding_error(ramona_runtime_errorbase):
	exitcode = 2

class server_start_error(ramona_runtime_errorbase):
	exitcode = 3

class configuration_error(ramona_runtime_errorbase):
	exitcode = 98

class parameters_error(ramona_runtime_errorbase):
	exitcode = 99



import sys, os, socket, signal, errno, weakref, logging, argparse, itertools, time
import pyev
from .. import cnscom
from ..config import config, read_config, config_files, config_includes, get_numeric_loglevel
from ..cnscom import program_state_enum, svrcall_error
from .cnscon import console_connection, message_yield_loghandler, deffered_return
from .proaster import program_roaster
from .idlework import idlework_appmixin
from .svrappsingl import server_app_singleton
from .notify import notificator

from . import call_status

###

L = logging.getLogger("server")

###

class server_app(program_roaster, idlework_appmixin, server_app_singleton):

	STOPSIGNALS = [signal.SIGINT, signal.SIGTERM]
	NONBLOCKING = frozenset([errno.EAGAIN, errno.EWOULDBLOCK])

	def __init__(self):
		server_app_singleton.__init__(self)

		# Create own process group
		os.setpgrp()

		# Parse command line arguments
		parser = argparse.ArgumentParser()
		self.args = parser.parse_args()

		# Read configuration
		read_config()

		# Configure logging
		loglvl = get_numeric_loglevel(config.get('ramona:server','loglevel'))
		logging.basicConfig(
			level=loglvl,
			stream=sys.stderr,
			format="%(asctime)s %(levelname)s: %(message)s",
			)
		L.debug("Configuration loaded from: {0}".format(', '.join(itertools.chain(config_files,config_includes))))

		# Prepare message yield logger
		my_logger = logging.getLogger('my')
		my_logger.setLevel(logging.DEBUG) 
		my_logger.addHandler(message_yield_loghandler(self))
		my_logger.propagate = False

		# Open console communication socket (listen mode)
		socket_factory = cnscom.socket_uri(config.get("ramona:server", "consoleuri"))
		try:
			self.sock = socket_factory.create_socket_listen()
		except socket.error, e:
			L.fatal("It looks like that server is already running: {0}".format(e))
			sys.exit(1)
		self.sock.setblocking(0)

		self.loop = pyev.default_loop()
		self.watchers = [pyev.Signal(sig, self.loop, self.__terminal_signal_cb) for sig in self.STOPSIGNALS]
		self.watchers.append(pyev.Child(0, False, self.loop, self.__child_signal_cb))
		self.watchers.append(pyev.Io(self.sock._sock, pyev.EV_READ, self.loop, self.__accept_cb))
		self.watchers.append(pyev.Periodic(0, 1.0, self.loop, self.__tick_cb))
		
		self.conns = weakref.WeakSet()
		self.termstatus =  None
		self.termstatus_change = None

		# Enable non-terminating SIGALARM handler
		signal.signal(signal.SIGALRM, _SIGALARM_handler)

		program_roaster.__init__(self)
		idlework_appmixin.__init__(self)

		# Build notificator component
		self.notificator = notificator(self)


	def run(self):
		self.sock.listen(socket.SOMAXCONN)
		for watcher in self.watchers:
			watcher.start()

		# Create pid file
		pidfile = config.get('ramona:server','pidfile')
		if pidfile !='':
			try:
				open(pidfile,'w').write("{0}\n".format(os.getpid()))
			except Exception, e:
				L.critical("Cannot create pidfile: {0}".format(e)) 
				del self.sock # Make sure that socket is explicitly closed (and eventual UNIX socket file deleted)
				sys.exit(1)
				
		# Launch loop
		try:
			self.loop.start()
		finally:
			# Close connections
			for conn in self.conns:
				conn.close()

			# Finalize idle work queue
			self.stop_idlework()

			# Finally remove pid file
			if pidfile !='':
				try:
					os.unlink(pidfile)
				except Exception, e:
					L.error("Cannot remove pidfile: {0}".format(e))
					self.sock.close()
					del self.sock # Make sure that socket is explicitly closed (and eventual UNIX socket file deleted)
					sys.exit(1)

		sys.exit(0)


	def __accept_cb(self, watcher, revents):
		try:
			while True:
				try:
					sock, address = self.sock.accept()
				except socket.error as err:
					if err.args[0] in self.NONBLOCKING:
						break
					else:
						raise
				else:
					if self.termstatus is not None: self.sock.close() # Do not accept new connection when exiting
					if sock.family==socket.AF_UNIX and address=='': address = sock.getsockname()
					conn = console_connection(sock, address, self)
					self.conns.add(conn)

		except:
			L.exception("Exception during server socket accept:")


	def __terminal_signal_cb(self, watcher, _revents):
		if watcher.signum == signal.SIGINT:
			# Print ENTER when Ctrl-C is pressed
			print

		if self.termstatus is None:
			L.info("Exit request received (by signal {0})".format(watcher.signum))
			self.__init_soft_exit()
			return

		else:
			self.__init_real_exit()
			return


	def __child_signal_cb(self, watcher, _revents):
		try:
			self.on_terminate_program(watcher.rpid, watcher.rstatus)
			self.add_idlework(self.on_tick) # Schedule extra periodic check 
		except:
			L.exception("Exception during SIGCHLD callback")


	def __tick_cb(self, watcher, revents):
		try:
			self.on_tick()
		except:
			L.exception("Exception during periodic internal check")


	def stop_loop(self):
		'''
		Stop internal loop and exit.
		'''
		self.loop.stop(pyev.EVBREAK_ALL)
		self.sock.close()
		while self.watchers:
			self.watchers.pop().stop()


	def dispatch_svrcall(self, cnscon, callid, params):
		if self.termstatus is not None:
			raise cnscom.svrcall_error('Ramona server is exiting - no further commands will be accepted')

		if callid == cnscom.callid_start:
			kwargs = cnscom.parse_json_kwargs(params)
			immediate = kwargs.pop('immediate', False)
			if immediate:
				return self.start_program(cnscon=None, **kwargs)
			else:
				cnscon.yield_enabled=True
				self.start_program(cnscon=cnscon, **kwargs)
				return deffered_return

		elif callid == cnscom.callid_stop:
			kwargs = cnscom.parse_json_kwargs(params)
			immediate = kwargs.pop('immediate', False)
			mode = kwargs.pop('mode',None)

			if mode is None or mode == 'stay':
				self.add_idlework(self.on_tick) # Schedule extra periodic check (to provide swift server background response to to user action)
				if immediate:
					return self.stop_program(cnscon=None, **kwargs)
				else:
					cnscon.yield_enabled=True
					self.stop_program(cnscon=cnscon, **kwargs)
					return deffered_return

			elif mode == 'exit':
				if immediate:
					return self.__init_soft_exit(cnscon=None, **kwargs)
				else:
					cnscon.yield_enabled=True
					self.__init_soft_exit(cnscon=cnscon, **kwargs)
					return deffered_return

			else:
				L.warning("Unknown exit mode issued: {0}".format(mode))

		elif callid == cnscom.callid_restart:
			self.add_idlework(self.on_tick) # Schedule extra periodic check (to provide swift server background response to to user action)
			kwargs = cnscom.parse_json_kwargs(params)
			immediate = kwargs.pop('immediate', False)
			if immediate:
				return self.restart_program(cnscon=None, **kwargs)
			else:
				cnscon.yield_enabled=True
				self.restart_program(cnscon=cnscon, **kwargs)
				return deffered_return

		elif callid == cnscom.callid_status:
			return call_status.main(self, **cnscom.parse_json_kwargs(params))

		elif callid == cnscom.callid_ping:
			return params

		elif callid == cnscom.callid_tail:
			kwargs = cnscom.parse_json_kwargs(params)
			program = kwargs.pop('program')
			try:
				program = self.get_program(program)
			except KeyError, e:
				raise svrcall_error("{0}".format(e.message))
			
			return program.tail(**kwargs)

		else:
			L.error("Received unknown callid: {0}".format(callid))


	def on_tick(self):
		now = time.time()
		program_roaster.on_tick(self, now)

		# If termination status take too long - do hard kill
		if self.termstatus_change is not None:
			if (now - self.termstatus_change) > 5:
				if self.termstatus != 3:
					self.__init_real_exit()
				else:
					L.fatal("It looks like server shutdown is taking way too much time - taking nasty exit")
					os._exit(10) 

		if (self.termstatus == 1) and (self.stop_seq is None):
			# Special care for server terminating condition 
			not_running_states=frozenset([program_state_enum.STOPPED, program_state_enum.FATAL, program_state_enum.CFGERROR, program_state_enum.DISABLED])
			ready_to_stop = True
			for p in self.roaster: # Seek for running programs
				if p.state not in not_running_states:
					ready_to_stop = False
					break

			if ready_to_stop: # Happy-flow (stop sequence finished and there is no program running - we can stop looping and exit)
				for p in self.roaster:
					if p.state in (program_state_enum.FATAL, program_state_enum.CFGERROR):
						L.warning("Process in error condition during exit: {0}".format(p))

				self.__init_soft2_exit(self)
			else:
				L.warning("Restarting stop sequence due to exit request.")
				self.stop_program(force=True)

		if (self.termstatus == 2):
			self.__close_idle_conns()
			if len(self.conns) == 0:
				self.__init_real_exit()


	def __init_soft_exit(self, cnscon=None, **kwargs):
		if self.termstatus > 1: return

		self.termstatus =  1
		self.termstatus_change = time.time()
		self.stop_program(cnscon=cnscon, force=True, **kwargs)
		self.add_idlework(self.on_tick) # Schedule extra periodic check 


	def __init_soft2_exit(self, cnscon=None):
		'''Term status 2: Clean idling console connections and wait for others console connections to close'''
		if self.termstatus > 2: return
		if self.termstatus == 1: self.__close_idle_conns()
		self.termstatus = 2
		self.termstatus_change = time.time()


	def __init_real_exit(self):
		self.termstatus = 3
		self.termstatus_change = time.time()
		self.stop_loop()


	def __close_idle_conns(self):
		for conn in list(self.conns):
			if conn.write_buf is None \
			   and len(conn.read_buf) == 0 \
			   and conn.yield_enabled is False \
			   and conn.return_expected is False:
				conn.close()


def _SIGALARM_handler(signum, frame):
	pass



import re, collections, io, os, glob, logging
from ..config import config
from ..kmpsearch import kmp_search
from .svrappsingl import get_svrapp

###

L = logging.getLogger('logmed')

###


class log_mediator(object):
	'''
	This object serves as mediator between program and its log files.

	It provides following functionality:
		- log rotation
		- tail buffer (TODO)
		- seek for patterns in log stream and eventually trigger error mail
	'''

	maxtailbuflen = 64*1024 # 64Kb is max len. of tail buffer
	rotlognamerg = re.compile('\.([0-9]+)$')

	def __init__(self, prog_ident, stream_name, fname):
		'''
		@param prog_ident: identification of the program (x from [program:x])
		@param stream_name: stdout or stderr
		@param fname: name of connected log file, can be None if no log is connected
		'''
		self.prog_ident = prog_ident
		self.stream_name = stream_name
		self.fname = fname
		self.outf = None
		self.scanners = []

		self.tailbuf = collections.deque()
		self.tailbuflen = 0

		# Read last content of the file into tail buffer
		if self.fname is not None and os.path.isfile(self.fname):
			with io.open(self.fname, 'r') as logf:
				if logf.seekable(): 
					logf.seek(0, io.SEEK_END)
					d = max(logf.tell() - self.maxtailbuflen, 0)
					logf.seek(d, io.SEEK_SET) # Seek to tail start position (end of file - maxtailbuflen)
					while True: # Read line by line into tail buffer
						data = logf.readline(4096)
						datalen = len(data)
						if datalen == 0: break
						self.__add_to_tailbuf(data)

		# Configure log rotation
		try:
			self.logmaxsize = config.getint('general','logmaxsize')
			self.logbackups = config.getint('general','logbackups')
		except Exception, e:
			self.logbackups = self.logmaxsize = 0
			L.warning("Invalid configuration of log rotation: {0} - log rotation disabled".format(e))


	def open(self):
		if self.outf is None and self.fname is not None:
			self.outf = open(self.fname,'a')
			

	def close(self):
		if self.outf is not None:
			self.outf.close()
			self.outf = None


	def write(self, data):
		if self.outf is not None:
			self.outf.write(data)
			self.outf.flush() #TODO: Maybe something more clever here can be better (check logging.StreamHandler)
			if (self.logmaxsize > 0) and (self.outf.tell() >= self.logmaxsize):
				self.rotate()

		self.__add_to_tailbuf(data)

		# Search for patterns
		if len(self.scanners) > 0:
			stext = data.lower()
			for s in self.scanners:
				r = s.search(stext)
				if r < 0: continue

				# Take last three tail entries (very likely lines)
				tail = ""
				for i in range(-3,0):
					try:
						tail += self.tailbuf[i][0]
					except IndexError:
						pass
				tail = tail[-2048:] # Limit result to 2kb of text

				svrapp = get_svrapp()
				if svrapp is not None:
					svrapp.notificator.publish(
						s.target,
						s.prog_ident,
						s.stream_name,
						''.join(s.pattern),
						tail
					)



	def rotate(self):
		'Perform rotation of connected file - if any'
		if self.fname is None: return
		if self.outf is None: return
		L.debug("Rotating '{0}' file".format(self.fname))

		self.outf.close()
		try:

			fnames = set()
			for fname in glob.iglob(self.fname+'.*'):
				if not os.path.isfile(fname): continue
				x = self.rotlognamerg.search(fname)
				if x is None: continue
				fnames.add(int(x.group(1)))

			for k in sorted(fnames, reverse=True):
				if (self.logbackups > 0) and (k >= self.logbackups):
					os.unlink("{0}.{1}".format(self.fname, k))
					continue
				if ((k-1) not in fnames) and (k > 1): continue # Move only files where there is one 'bellow'
				os.rename("{0}.{1}".format(self.fname, k), "{0}.{1}".format(self.fname, k+1))

			os.rename("{0}".format(self.fname), "{0}.1".format(self.fname))

		finally:
			self.outf = open(self.fname,'a')		


	def __add_to_tailbuf(self, data):
		# Add data to tail buffer
		datalen = len(data)
		self.tailbuf.append((data, datalen))
		self.tailbuflen += datalen

		# Clean tail buffer - data that exceeds max. length
		while self.tailbuflen > self.maxtailbuflen:
			try:
				_, odatalen = self.tailbuf.popleft()
			except IndexError:
				self.tailbuflen = 0
				break

			self.tailbuflen -= odatalen


	def tail(self):
		d = collections.deque()
		dlen = 0
		for data, datalen in reversed(self.tailbuf):
			dlen += datalen
			if dlen >= 0x7fff: break #Protect maximum IPC data len
			d.appendleft(data)

		return "".join(d)


	def add_scanner(self, pattern, target):
		self.scanners.append(
			_log_scanner(self.prog_ident, self.stream_name, pattern, target)
		)


class _log_scanner(kmp_search):

	def __init__(self, prog_ident, stream_name, pattern, target):
		kmp_search.__init__(self, pattern)
		assert target in ('now','daily') or target.startswith('mailto:')
		self.target = target
		self.prog_ident = prog_ident
		self.stream_name = stream_name




class kmp_search(object):
	'''Implementation of Knuth-Morris-Pratt string matching algorithm
	See http://en.wikipedia.org/wiki/Knuth-Morris-Pratt_algorithm for more details.

	Basically this is useful for searching pattern (string) in the non-persistent stream of data.
	'''

	def __init__(self, pattern):
		# allow indexing into pattern and protect against change during yield
		self.pattern = list(pattern)
		self.patternlen = len(self.pattern)

		# build table of shift amounts
		self.shifts = [1] * (self.patternlen + 1)
		shift = 1
		for pos in range(self.patternlen):
			while shift <= pos and self.pattern[pos] != self.pattern[pos-shift]:
				shift += self.shifts[pos-shift]
		self.shifts[pos+1] = shift

		self.startpos = 0
		self.matchsel = 0


	def search(self, text):
		for c in text:
			while self.matchsel == self.patternlen or self.matchsel >= 0 and self.pattern[self.matchsel] != c:
				self.startpos += self.shifts[self.matchsel]
				self.matchsel -= self.shifts[self.matchsel]
			self.matchsel += 1
			if self.matchsel == self.patternlen:
				return self.startpos
		return -1




import json
from ... import cnscom
from .. import exception
from ._completes import complete_ident
###

name = 'tail'
cmdhelp = 'Tail log of specified program'

###

def init_parser(parser):
	parser.add_argument('-l','--log-stream', choices=['stdout','stderr'], default='stderr', help='Specify which log stream to use in tail (default is stderr)')
	parser.add_argument('program', help='Specify program(s) in scope of the command')

###

def complete(cnsapp, text, line, begidx, endidx):
	textst = text.strip()
	ret = []
	ret.extend(complete_ident(cnsapp, textst))
	return ret

###


def main(cnsapp, args):

	params = {
		'program': args.program,
		'stream': args.log_stream,
	}
	ret = cnsapp.svrcall(
		cnscom.callid_tail,
		json.dumps(params),
		auto_connect=True
	)

	print ret



import logging, time
from ..config import config
from ..cnscom import svrcall_error, program_state_enum
from .program import program
from .seqctrl import sequence_controller

###

L = logging.getLogger('proaster')
Lmy = logging.getLogger('my')

###

class program_roaster(object):
	'''
Program roaster is object that control all configured programs, their start/stop operations etc.
	'''

	def __init__(self):
		self.start_seq = None
		self.stop_seq = None
		self.restart_seq = None

		self.roaster = []
		for config_section in config.sections():
			if config_section.find('program:') != 0: continue
			sp = program(self, config_section)
			self.roaster.append(sp)


	def get_program(self, ident):
		for p in self.roaster:
			if p.ident == ident: return p
		raise KeyError("Unknown program '{0}' requested".format(ident))


	def filter_roaster_iter(self, pfilter=None):
		if pfilter is None:
			for p in self.roaster: yield p
			return

		filter_set = frozenset(pfilter)
		roaster_dict = dict((p.ident, p) for p in self.roaster)

		# Pass only known program names
		not_found = filter_set.difference(roaster_dict)
		if len(not_found) > 0:
			for pn in not_found:
				Lmy.error('Unknown/invalid program name: {0}'.format(pn))

		for ident, p in roaster_dict.iteritems():
			if ident in filter_set: yield p


	def start_program(self, cnscon, pfilter=None, force=False):
		'''Start processes that are STOPPED and (forced) FATAL'''
		if self.start_seq is not None or self.stop_seq is not None or self.restart_seq is not None:
			raise svrcall_error("There is already start/stop sequence running - please wait and try again later.")

		l = self.filter_roaster_iter(pfilter)

		L.debug("Initializing start sequence")
		self.start_seq = sequence_controller(cnscon)

		# If 'force' is used, include as programs in FATAL state
		if force: states = (program_state_enum.STOPPED,program_state_enum.FATAL)
		else: states = (program_state_enum.STOPPED,)

		for p in l:
			if p.state in states:
				self.start_seq.add(p)		
			else:
				Lmy.warning("Program {0} is in {1} state - not starting.".format(p.ident, program_state_enum.labels.get(p.state,'<?>')))

		self.__startstop_pad_next(True)


	def stop_program(self, cnscon, pfilter=None, force=False, coredump=False):
		'''
		Stop processes that are RUNNING and STARTING
		@param force: If True then it interrupts any concurrently running start/stop sequence.
		'''
		if force:
			self.start_seq = None
			self.restart_seq = None
			self.stop_seq = None

		else:
			if self.start_seq is not None or self.stop_seq is not None or self.restart_seq is not None:
				raise svrcall_error("There is already start/stop sequence running - please wait and try again later.")

		l = self.filter_roaster_iter(pfilter)

		L.debug("Initializing stop sequence")
		self.stop_seq = sequence_controller(cnscon)

		for p in l:
			if p.state not in (program_state_enum.RUNNING, program_state_enum.STARTING): continue
			if coredump: p.charge_coredump()
			self.stop_seq.add(p)

		self.__startstop_pad_next(False)


	def restart_program(self, cnscon, pfilter=None, force=False):
		'''Restart processes that are RUNNING, STARTING, STOPPED and (forced) FATAL'''
		if self.start_seq is not None or self.stop_seq is not None or self.restart_seq is not None:
			raise svrcall_error("There is already start/stop sequence running - please wait and try again later.")

		L.debug("Initializing restart sequence")
		
		l = self.filter_roaster_iter(pfilter)

		self.stop_seq = sequence_controller() # Don't need to have cnscon connected with stop_seq (there is no return)
		self.restart_seq = sequence_controller(cnscon)

		# If 'force' is used, include as programs in FATAL state
		if force: start_states = (program_state_enum.STOPPED,program_state_enum.FATAL)
		else: start_states = (program_state_enum.STOPPED,)

		for p in l:
			if p.state in (program_state_enum.RUNNING, program_state_enum.STARTING):
				self.stop_seq.add(p)
				self.restart_seq.add(p)
			elif p.state in start_states:
				self.restart_seq.add(p)
			else:
				Lmy.warning("Program {0} is in {1} state - not restarting.".format(p.ident, program_state_enum.labels.get(p.state,'<?>')))

		self.__startstop_pad_next(False)



	def __startstop_pad_next(self, start=True):
		pg = self.start_seq.next() if start else self.stop_seq.next()
		if pg is None:
			if start:
				cnscon = self.start_seq.cnscon
				if cnscon is not None:
					self.start_seq.cnscon = None
					cnscon.send_return(True)
				self.start_seq = None
				L.debug("Start sequence completed.")
			else:
				cnscon = self.stop_seq.cnscon

				if self.restart_seq is None or self.termstatus is not None:
					if cnscon is not None:
						self.stop_seq.cnscon = None
						cnscon.send_return(True)
					L.debug("Stop sequence completed.")
					self.stop_seq = None
					return

				else:
					Lmy.info("Restart finished stop phase and entering start phase")
					L.debug("Restart sequence enters starting phase")
					self.stop_seq = None
					self.start_seq = self.restart_seq
					self.restart_seq = None
					self.__startstop_pad_next(True)
					return

		else:
			# Start/stop all programs in the active set
			map(program.start if start else program.stop, pg)


	def on_terminate_program(self, pid, status):
		for p in self.roaster:
			if pid != p.pid: continue
			return p.on_terminate(status)
		else:
			L.warning("Unknown program died (pid={0}, status={1})".format(pid, status))


	def on_tick(self, now):
		'''Periodic check of program states'''
		for p in self.roaster:
			p.on_tick(now)

		if self.start_seq is not None:
			r = self.start_seq.check(program_state_enum.STARTING, program_state_enum.RUNNING)
			if r is None:
				L.warning("Start sequence aborted due to program error")
				self.start_seq = None
				assert self.restart_seq is None
			elif r:
				self.__startstop_pad_next(True)

		if self.stop_seq is not None:
			r = self.stop_seq.check(program_state_enum.STOPPING, program_state_enum.STOPPED)
			if r is None:
				if self.restart_seq is None:
					L.warning("Stop sequence aborted due to program error")
					self.stop_seq = None
					assert self.start_seq is None
					assert self.restart_seq is None
				else:
					L.warning("Restart sequence aborted due to program error")
					self.restart_seq = None
					self.stop_seq = None
					assert self.start_seq is None

			elif r:
				self.__startstop_pad_next(False)



import sys, os, socket, ConfigParser, errno, logging, httplib, BaseHTTPServer, mimetypes, json, signal
import time, cgi, pprint, urllib, urlparse, itertools, base64, hashlib
from ..config import config, read_config, get_numeric_loglevel
from .. import cnscom

###

L = logging.getLogger("httpfendapp")

###

STRFTIME_FMT = "%d-%b-%Y %H:%M:%S"

###

# Initialize mimetypes
if not mimetypes.inited:
		mimetypes.init()

###

class httpfend_app(object):

	instance = None

	def __init__(self):
		assert self.instance is None
		httpfend_app.instance = self
		
		# Read config
		read_config()
		
		# Configure logging
		try:
			loglvl = get_numeric_loglevel(config.get(os.environ['RAMONA_SECTION'], 'loglevel'))
		except:
			loglvl = logging.INFO
		logging.basicConfig(
			level=loglvl,
			stream=sys.stderr,
			format="%(asctime)s %(levelname)s: %(message)s",
		)

		try:
			host = config.get(os.environ['RAMONA_SECTION'], 'host')
		except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
			host = "localhost"
		
		try:
			port = config.getint(os.environ['RAMONA_SECTION'], 'port')
		except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
			port = 5588
		
		self.username = None
		self.password = None 
		try:
			self.username = config.get(os.environ['RAMONA_SECTION'], 'username')
			self.password = config.get(os.environ['RAMONA_SECTION'], 'password')
		except:
			pass
		
		if self.username is not None and self.password is None:
			L.fatal("Configuration error: 'username' option is set, but 'password' option is not set. Please set 'password'")
			sys.exit(1) 

		Handler = RamonaHttpReqHandler
		self.httpd = BaseHTTPServer.HTTPServer((host, port), Handler)
		
		# Prepare server connection factory
		self.cnsconuri = cnscom.socket_uri(config.get('ramona:console','serveruri'))
		
		self.logmsgcnt = itertools.count()
		self.logmsgs = dict()
		signal.signal(signal.SIGINT, self.stop)

	def run(self):
		L.info("Started HTTP frontend at http://{0}:{1}".format(self.httpd.server_name, self.httpd.server_port))
		self.httpd.serve_forever()
	
	def stop(self, signum, frame):
		L.info("Received signal {0}. Stopping the server.".format(signum))
		self.httpd.shutdown()


class RamonaHttpReqHandler(BaseHTTPServer.BaseHTTPRequestHandler):
	
	ActionToCallid = {"start": cnscom.callid_start, "stop": cnscom.callid_stop, "restart": cnscom.callid_restart}
	scriptdir = os.path.join(".", "ramona", "httpfend")
	
	def do_GET(self):
		
		if self.path.startswith("/static/"):
			parts = self.path.split("/")
			fname = os.path.join(self.scriptdir, *[x for x in parts if len(x) > 0])
			if not os.path.isfile(fname):
				self.send_error(httplib.NOT_FOUND)
				return
			try:
				f = open(fname, "r")
			except IOError:
				self.send_error(httplib.NOT_FOUND)
				return
			with f:
				self.send_response(httplib.OK)
				self.send_header("Content-Type", mimetypes.guess_type(self.path)[0])
				self.end_headers()
				self.wfile.write(f.read())
				return
		
		authheader = self.headers.getheader("Authorization", None)
		if httpfend_app.instance.username is not None and authheader is None:
			self.serve_auth_headers()
			return
		
		elif httpfend_app.instance.username is not None and authheader is not None:
			method, authdata = authheader.split(" ") 
			if method != "Basic":
				self.send_error(httplib.NOT_IMPLEMENTED, "The authentication method '{0}' is not supported. Only Basic authnetication method is supported.".format(method))
				return
			username, _, password = base64.b64decode(authdata).partition(":")
			if httpfend_app.instance.password.startswith("{SHA}"):
				password = "{SHA}" + hashlib.sha1(password).hexdigest()
			
			if username != httpfend_app.instance.username or password != httpfend_app.instance.password:
				self.serve_auth_headers()
				return
		
		if self.path.startswith("/ajax/"):
			parsed = urlparse.urlparse(self.path)
			action = parsed.path[6:]
			if action == "statusTable":
				self.send_response(httplib.OK)
				self.send_header("Content-Type", "text/html; charset=utf-8")
				self.end_headers()
				self.wfile.write(self.buildStatusTable(json.loads(self.getStatuses())))
		
		elif self.path.startswith("/log/"):
			parsed = urlparse.urlparse(self.path)
			logname = parsed.path[5:]
			parts = logname.split("/")
			if len(parts) < 2:
				self.send_error(httplib.NOT_FOUND, "Invalid URL.")
				return
			stream = parts[0]
			if stream not in ("stdout", "stderr"):
				self.send_error(httplib.NOT_FOUND, "'{0}' is not a valid type of stream. Only 'stdout' and 'stderr' are supported.".format(stream))
				return
			program = urllib.unquote_plus(parts[1].rpartition(".")[0])
			conn = self.socket_connect()
			params = {
					"program": program,
					"stream": stream, 
			}
			try:
				ret = cnscom.svrcall(conn, cnscom.callid_tail, json.dumps(params))
				self.send_response(httplib.OK)
				self.send_header("Content-Type", "text/plain; charset=utf-8")
				self.end_headers()
				self.wfile.write(ret)
			except Exception, e:
				self.send_error(httplib.INTERNAL_SERVER_ERROR, str(e))
				return

		else:
			parsed = urlparse.urlparse(self.path)
			if parsed.path != "/":
				self.send_error(httplib.NOT_FOUND)
				return
			
			qs = urlparse.parse_qs(parsed.query)
			action = None
			actionList = qs.get('action')
			if actionList is not None and len(actionList) > 0:
				action = actionList[0]
			if action in ("start", "stop", "restart"):
				conn = self.socket_connect()
				params = {
					'immediate': True,
				}
				qsIdent = qs.get('ident')
				if qsIdent is not None and len(qsIdent) > 0:
					params['pfilter'] = [qsIdent[0]]
				else:
					params['pfilter'] = list(self.getAllPrograms(True))
				
				qsForce = qs.get('force')
				if qsForce is not None and len(qsForce) > 0:
					if qsForce[0] == "1":
						params['force'] = True
				
				try:
					cnscom.svrcall(conn, self.ActionToCallid[action], json.dumps(params))
					msgid = self.addLogMessage("success", "Command '{0}' successfully triggered.".format(action))
				except Exception, e:
					msgid = self.addLogMessage("error", "Failed to trigger the command: {0}".format(e))
				
				self.send_response(httplib.SEE_OTHER)
				self.send_header("Location", self.getAbsPath(msgid=msgid))
				self.end_headers()
				return
			
			self.send_response(httplib.OK)
			self.send_header("Content-Type", "text/html; charset=utf-8")
			self.end_headers()
			
			logmsg = ""
			for msgid in qs.get('msgid', []):
				m = httpfend_app.instance.logmsgs.pop(int(msgid), None)
				if m is not None:
					logmsg += '''<div class="alert alert-{0}">{1}</div>'''.format(*m)

			with open(os.path.join(self.scriptdir, "index.tmpl.html")) as f:
				sttable = self.buildStatusTable(json.loads(self.getStatuses()))
				self.wfile.write(f.read().format(
					statuses=sttable,
					logmsg=logmsg,
					appname=config.get('general','appname')
				))
				
	def log_message(self, fmt, *args):
		L.debug("{0} -- [{1}]: {2}".format(self.address_string(), self.log_date_time_string(), fmt % args))
			
	
	def buildStatusTable(self, statuses):
		ret = '<table id="statusTable" class="table table-hover table-bordered"><thead>'
		ret += '<tr><th>Process ident.</th><th>Status</th><th><abbr title="O = stdout, E = stderr">Log</abbr></th><th>PID</th><th><abbr title="Launch counter">L.cnt.</abbr></th><th>Start time</th><th>Exit time</th><th>Exit code</th><th></th></tr>'
		ret += "</thead>"
		ret += "<tbody>"
		for sp in statuses:
			ret += "<tr>"
			ident = sp.pop('ident', '???')
			ret += '<th>{0}</th>'.format(cgi.escape(ident))
			labelCls = "label-inverse"
			progState = sp.pop("state")
			
			if progState == cnscom.program_state_enum.RUNNING:
				labelCls = "label-success"
			elif progState == cnscom.program_state_enum.STOPPED:
				labelCls = ""
			elif progState in (cnscom.program_state_enum.STOPPING, cnscom.program_state_enum.STARTING):
				labelCls = "label-info"
			elif progState == cnscom.program_state_enum.STOPPED:
				labelCls = ""
			elif progState in (cnscom.program_state_enum.FATAL, cnscom.program_state_enum.CFGERROR):
				labelCls = "label-important"
			
			stlbl = cnscom.program_state_enum.labels.get(progState, "({0})".format(progState))
			ret += '<td><span class="label {0}">{1}</span></td>'.format(labelCls, cgi.escape(stlbl))
			ret += '<td><a href="/log/stdout/{0}.log" target="_blank">O</a> <a href="/log/stderr/{0}.log" target="_blank">E</a>'.format(urllib.quote_plus(ident))
			pid = sp.pop('pid', "")
			ret += '<td>{0}</td>'.format(pid)
			ret += '<td>{0}</td>'.format(sp.pop('launch_cnt', ""))

			u = sp.pop('uptime', None)
			if u is not None:
				t = sp.pop('start_time', '?')
				ret += '<td title="Started at {1}">{0}</td>'.format(natural_relative_time(u), time.strftime(STRFTIME_FMT,time.localtime(t)))
			else:
				t = sp.pop('start_time', None)
				tform = ""
				if t is not None: tform = time.strftime(STRFTIME_FMT,time.localtime(t))
				ret += '<td>{0}</td>'.format(tform)

			t = sp.pop('exit_time', None)
			tform = ""
			if t is not None: tform = time.strftime(STRFTIME_FMT,time.localtime(t))
			ret += '<td>{0}</td>'.format(tform)
			ret += '<td>{0}</td>'.format(sp.pop('exit_status',''))
			
			actions = []
			if pid != os.getpid():
				# TODO: Should there be some filtering for STOPPING ???
				if progState not in (cnscom.program_state_enum.FATAL, cnscom.program_state_enum.RUNNING, cnscom.program_state_enum.STARTING, cnscom.program_state_enum.DISABLED):
					actions.append('<a class="btn btn-small btn-success" href="/?{0}">Start</a>'.format(urllib.urlencode([("action", "start"), ("ident", ident)])))
				
				if progState == cnscom.program_state_enum.RUNNING:
					actions.append('<a class="btn btn-small btn-danger" href="/?{0}">Stop</a>'.format(urllib.urlencode([("action", "stop"), ("ident", ident)])))
					actions.append('<a class="btn btn-small btn-warning" href="/?{0}">Restart</a>'.format(urllib.urlencode([("action", "restart"), ("ident", ident)])))
			
				if progState == cnscom.program_state_enum.FATAL:
					actions.append('<a class="btn btn-small btn-inverse" href="/?{0}">Start (force)</a>'.format(urllib.urlencode([("action", "start"), ("ident", ident), ("force", "1")])))

			ret += '<td>{0}</td>'.format(" ".join(actions))
			ret += "</tr>"
			
			if len(sp) > 0:
				ret += '<tr class="info"><td colspan="2"></td><td colspan="7"><pre class="pre-scrollable">'
				ret += cgi.escape(pprint.pformat(sp, width=3))
				ret += '</pre></td></tr>'
		
		ret += "</tbody></table>"
		return ret
	
	
	def addLogMessage(self, level, msg):
		msgid = httpfend_app.instance.logmsgcnt.next()
		httpfend_app.instance.logmsgs[msgid] = (level, msg)
		return msgid
	
	def getAbsPath(self, path="/", **kwargs):
		queryList = []
		for k,v in kwargs.iteritems():
			queryList.append((k, v))
		
		return urlparse.urlunparse(("http", self.headers['Host'], "/", None, urllib.urlencode(queryList), None))
	
	def getStatuses(self):
		conn = self.socket_connect()
		return cnscom.svrcall(conn, cnscom.callid_status, json.dumps({}))
	
	def getAllPrograms(self, withoutSelf=False):
		'''
		@param withoutSelf: If true, the process with the same pid will be ignored
		@return iterator: Ident of all registered programs
		'''
		for st in json.loads(self.getStatuses()):
			ident = st.get("ident")
			pid = st.get("pid")
			if ident is None:
				continue
			if (not withoutSelf) or (pid != os.getpid()):
				yield ident
		
	
	def socket_connect(self):
		if hasattr(self, "socket_conn"): return self.socket_conn
		try:
			self.socket_conn = httpfend_app.instance.cnsconuri.create_socket_connect()
			return self.socket_conn
		except socket.error, e:
			if e.errno == errno.ECONNREFUSED: return None
			if e.errno == errno.ENOENT and self.cnsconuri.protocol == 'unix': return None
			raise
	
	def serve_auth_headers(self):
		self.send_response(httplib.UNAUTHORIZED)
		self.send_header('WWW-Authenticate', 'Basic realm="Ramona HTTP frontend"')
		self.send_header('Content-type', 'text/html')
		self.end_headers()
		with open(os.path.join(self.scriptdir, "401.tmpl.html")) as f:
			self.wfile.write(f.read().format(
					appname=config.get('general','appname'),
					configsection=os.environ['RAMONA_SECTION'])
			)
		

def natural_relative_time(diff_sec):
	#TODO: Improve this significantly - maybe even add unit test
	#if diff.days > 7 or diff.days < 0: return d.strftime('%d %b %y')
#	elif diff.days == 1:
#	    return '1 day ago'
#	elif diff.days > 1:
#	    return '{} days ago'.format(diff.days)
	if diff_sec <= 1:
		return 'just now'
	elif diff_sec < 60:
		return '{:0.1f} sec(s) ago'.format(diff_sec)
	elif diff_sec < 120:
		return '1 min ago'
	elif diff_sec < 3600:
		return '{:0.0f} min(s) ago'.format(diff_sec/60)
	elif diff_sec < 7200:
		return '1 hour ago'
	else:
		return '{:0.0f} hours ago'.format(diff_sec/3600)



import os, urlparse, socket, struct, time, json, select, logging
###

L = logging.getLogger("cnscom")
Lmy = logging.getLogger("my")

###

callid_ping = 0
callid_start = 1
callid_stop = 2
callid_restart = 3
callid_status = 4
callid_tail = 5

#

call_magic = '>'
resp_magic = '<'

call_struct_fmt = '!cBH'
resp_struct_fmt = '!ccH'

resp_return = 'R'
resp_exception = 'E'
resp_yield_message = 'M' # Used to propagate message from server to console

###

class program_state_enum(object):
	'''Enum'''
	DISABLED = -1
	STOPPED = 0
	STARTING = 10
	RUNNING = 20
	STOPPING = 30
	FATAL = 200
	CFGERROR=201

	labels = {
		DISABLED: 'DISABLED',
		STOPPED: 'STOPPED',
		STARTING: 'STARTING',
		RUNNING: 'RUNNING',
		STOPPING: 'STOPPING',
		FATAL: 'FATAL',
		CFGERROR: 'CFGERROR',
	}


###


def svrcall(cnssocket, callid, params=""):
	'''
	Client side of console communication IPC call (kind of RPC / Remote procedure call).

	@param cnssocket: Socket to server (created by socket_uri factory bellow)
	@param callid: one of callid_* identification
	@param params: string representing parameters that will be passed to server call
	@return: String returned by server or raises exception if server call failed
	'''

	paramlen = len(params)
	if paramlen >= 0x7fff:
		raise RuntimeError("Transmitted parameters are too long.")

	cnssocket.send(struct.pack(call_struct_fmt, call_magic, callid, paramlen)+params)

	while 1:
		x = time.time()
		resp = ""
		while len(resp) < 4:
			rlist, _, _ = select.select([cnssocket],[],[], 5)
			if len(rlist) == 0:
				if time.time() - x > 2: L.error("Looping detected")
				continue
			ndata = cnssocket.recv(4 - len(resp))
			if len(ndata) == 0:
				raise EOFError("It looks like server closed connection")

			resp += ndata

		magic, retype, paramlen = struct.unpack(resp_struct_fmt, resp)
		assert magic == resp_magic

		# Read rest of the response (size given by paramlen)
		params = ""
		while paramlen > 0:
			ndata = cnssocket.recv(paramlen)
			params += ndata
			paramlen -= len(ndata)


		if retype == resp_return:
			# Remote server call returned normally
			return params
		
		elif retype == resp_exception:
			# Remove server call returned exception
			raise RuntimeError(params)
		
		elif retype == resp_yield_message:
			# Remote server call returned yielded message -> we will continue receiving
			obj = json.loads(params)
			obj = logging.makeLogRecord(obj)
			if Lmy.getEffectiveLevel() <= obj.levelno: # Print only if log level allows that
				Lmy.handle(obj)
			continue

		else:
			raise RuntimeError("Unknown server response: {0}".format(retype))

###

class socket_uri(object):
	'''
	Socket factory that is configured using socket URI.
	This is actually quite generic implementation - not specific to console-server IPC communication.
	'''

	# Configure urlparce
	if 'unix' not in urlparse.uses_params: urlparse.uses_params.append('unix')

	def __init__(self, uri):
		self.uri = urlparse.urlparse(uri)
		self.uriparams = dict(urlparse.parse_qsl(self.uri.params))

		self.protocol = self.uri.scheme.lower()
		if self.protocol == 'tcp':
			try:
				_port = self.uri.port
			except ValueError:
				raise RuntimeError("Invalid port number in socket URI {0}".format(uri))

			if self.uri.path != '': raise RuntimeError("Path has to be empty in socket URI {0}".format(uri))

		elif self.protocol == 'unix':
			if self.uri.netloc != '':
				# Special case of situation when netloc is not empty (path is relative)
				self.uri = self.uri._replace(netloc='', path=self.uri.netloc + self.uri.path)

		else:
			raise RuntimeError("Unknown/unsuported protocol '{0}' in socket URI {1}".format(self.protocol, uri))


	def create_socket_listen(self):
		if self.protocol == 'tcp':
			s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
			s.bind((self.uri.hostname, self.uri.port))

		elif self.protocol == 'unix':
			mode = self.uriparams.get('mode',None)
			if mode is None: mode = 0o600
			else: mode = int(mode,8)
			oldmask = os.umask(mode ^ 0o777)
			s = deleteing_unix_socket()
			s.bind(self.uri.path)
			os.umask(oldmask)

		else:
			raise RuntimeError("Unknown/unsuported protocol '{0}'".format(self.protocol))
		
		return s


	def create_socket_connect(self):
		if self.protocol == 'tcp':
			s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			s.connect((self.uri.hostname, self.uri.port))

		elif self.protocol == 'unix':
			s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
			s.connect(self.uri.path)

		else:
			raise RuntimeError("Unknown/unsuported protocol '{0}'".format(self.protocol))
		
		return s

###

class deleteing_unix_socket(socket.socket):
	'''
This class is used as wrapper to socket object that represent listening UNIX socket.
It added ability to delete socket file when destroyed.

It is basically used only on server side of UNIX socket.
	'''

	def __init__(self):
		socket.socket.__init__(self, socket.AF_UNIX, socket.SOCK_STREAM)
		self.__sockfile = None


	def __del__(self):
		self.__delsockfile()


	def close(self):
		socket.socket.close(self)
		self.__delsockfile()


	def bind(self, fname):
		socket.socket.bind(self, fname)
		self.__sockfile = fname


	def __delsockfile(self):
		if self.__sockfile is not None:
			fname = self.__sockfile
			self.__sockfile = None
			os.unlink(fname)
			assert not os.path.isfile(fname)


###

def parse_json_kwargs(params):
	'''Used when params are transfered as JSON - it also handles situation when 'params' is empty string '''
	if params == '': return dict()
	return json.loads(params)

###

class svrcall_error(RuntimeError):
	'''
	Exception used to report error to the console without leaving trace in server error log.
	'''
	pass
